from collections.abc import Callable
from functools import partial
from typing import Union

import jax
import jax.numpy as jnp
from jax.flatten_util import ravel_pytree

import netket.jax as nkjax
from netket.utils import timing
from netket.utils.types import Array, PyTree
from netket.vqs import FullSumState
from netket._src.ngd.sr import _compute_sr_update
from netket._src.ngd.srt import _compute_srt_update


def get_samples_and_pdf(vstate):
    if isinstance(vstate, FullSumState):
        samples = vstate.hilbert.all_states()
        pdf = vstate.probability_distribution()
    else:
        samples = vstate.samples
        samples = jax.lax.collapse(samples, 0, samples.ndim - 1)
        pdf = None
    return samples, pdf


def _prepare_weights(
    weights: Union[Array, None],
    n_samples: int,
) -> tuple[Union[Array, None], Union[Array, float]]:
    if weights is not None:
        # Normalize weights for self-normalized importance sampling
        weights = weights / jnp.mean(weights)
        # p(x) = q(x) * w(x) / jnp.mean(w)
        pdf = weights / n_samples
        mass = pdf
    else:
        pdf = None
        mass = 1 / n_samples
    return pdf, mass


@partial(jax.jit, static_argnames=("mode",))
def _prepare_input(
    O_L,
    local_grad,
    *,
    mode: str,
    scaling_factor: Union[Array, float],
) -> tuple[jax.Array, jax.Array]:
    r"""
    Prepare the input for the SR/SRt solvers.

    The local energies and the jacobian are reshaped, centered and normalized by the number of Monte Carlo samples.
    The complex case is handled by concatenating the real and imaginary parts of the jacobian and the local energies.

    We use [Re_x1, Im_x1, Re_x2, Im_x2, ...] so that shards are contiguous, and jax can keep track of the sharding information.
    This format is applied both to the jacobian and to the vector.

    Args:
        O_L: The jacobian of the ansatz.
        local_grad: The local energies.
        mode: The mode of the jacobian: `'real'` or `'complex'`.
        scaling_factor: The mass given to each sample, e.g. `1/n_samples` when using Monte Carlo sampling.

    Returns:
        The reshaped jacobian and the reshaped local energies.
    """
    local_grad = local_grad.flatten()
    de = local_grad - jnp.sum(scaling_factor * local_grad)

    dv = 2.0 * de * jnp.sqrt(scaling_factor)
    if jnp.ndim(scaling_factor) != 0:
        scaling_factor = jax.lax.broadcast_in_dim(scaling_factor, O_L.shape, (0,))
    O_L = O_L * jnp.sqrt(scaling_factor)

    if mode == "complex":
        # Concatenate the real and imaginary derivatives of the ansatz
        # (#ns, 2, np) -> (#ns*2, np)
        O_L = jax.lax.collapse(O_L, 0, 2)

        # (#ns, 2) -> (#ns*2)
        dv2 = jnp.stack([jnp.real(dv), jnp.imag(dv)], axis=-1)
        dv = jax.lax.collapse(dv2, 0, 2)
    elif mode == "real":
        dv = dv.real
    else:
        raise NotImplementedError()
    return O_L, dv


@timing.timed
@partial(
    jax.jit,
    static_argnames=(
        "log_psi",
        "solver_fn",
        "mode",
        "chunk_size",
        "use_ntk",
    ),
)
def _sr_srt_common(
    log_psi,
    local_grad,
    parameters,
    model_state,
    samples,
    *,
    diag_shift: float | Array,
    solver_fn: Callable[[Array, Array], Array],
    mode: str,
    proj_reg: float | Array | None = None,
    momentum: float | Array | None = None,
    old_updates: PyTree | None = None,
    chunk_size: int | None = None,
    use_ntk: bool = False,
    weights: Array | None = None,
):
    r"""
    Compute the SR/Natural gradient update for the model specified by
    `log_psi({parameters, model_state}, samples)` and the local gradient contributions `local_grad`.

    Uses a code equivalent to QGTJacobianDense by default, or with the NTK/MinSR if `use_ntk` is True.

    Args:
        log_psi: The log of the wavefunction.
        local_grad: The local values of the estimator.
        parameters: The parameters of the model.
        model_state: The state of the model.
        diag_shift: The diagonal shift of the stochastic reconfiguration matrix. Typical values are 1e-4 ÷ 1e-3. Can also be an optax schedule.
        proj_reg: Weight before the matrix `1/N_samples \\bm{1} \\bm{1}^T` used to regularize the linear solver in SPRING.
        momentum: Momentum used to accumulate updates in SPRING.
        linear_solver: Callable to solve the linear problem associated to the updates of the parameters.
        mode: The mode used to compute the jacobian of the variational state. Can be `'real'` or `'complex'` (defaults to the dtype of the output of the model).
        weights: Importance sampling weights for the samples. If `None`, uniform weights are assumed.

    Returns:
        The new parameters, the old updates, and the info dictionary.
    """
    if use_ntk and weights is not None:
        raise NotImplementedError(
            "pdf is not currently supported when using NTK/MinSR. Please set pdf=None."
        )

    _, unravel_params_fn = ravel_pytree(parameters)
    _params_structure = jax.tree_util.tree_map(
        lambda x: jax.ShapeDtypeStruct(x.shape, x.dtype), parameters
    )

    pdf, mass = _prepare_weights(weights, samples.shape[0])

    jacobians = nkjax.jacobian(
        log_psi,
        parameters,
        samples,
        model_state,
        mode=mode,
        dense=True,
        center=True,
        chunk_size=chunk_size,
        pdf=pdf,
    )  # jacobian is centered

    O_L, dv = _prepare_input(jacobians, local_grad, mode=mode, scaling_factor=mass)

    if old_updates is None and momentum is not None:
        old_updates = jnp.zeros(jacobians.shape[-1], dtype=jacobians.dtype)

    compute_update = _compute_srt_update if use_ntk else _compute_sr_update

    # TODO: Add support for proj_reg and momentum
    # At the moment SR does not support momentum, proj_reg.
    # We raise an error if they are passed with a value different from None.
    updates, old_updates, info = compute_update(
        O_L,
        dv,
        diag_shift=diag_shift,
        solver_fn=solver_fn,
        mode=mode,
        proj_reg=proj_reg,
        momentum=momentum,
        old_updates=old_updates,
        params_structure=_params_structure,
    )

    return unravel_params_fn(updates), old_updates, info


sr = partial(_sr_srt_common, use_ntk=False)

srt = partial(_sr_srt_common, use_ntk=True)



# Copyright 2023 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

import jax

from flax import linen as nn

from netket.jax import logsumexp_cplx
from netket.utils.group import PermutationGroup
from netket.utils import HashableArray


class SymmExpSum(nn.Module):
    r"""
    A wrapper module to symmetrise a variational wave function with respect
    to a permutation group :math:`G` by summing over all permuted inputs.

    Given the characters :math:`\chi_g` of an irreducible representation
    (irrep) of :math:`G`, a variational state that transforms according
    to this irrep is given by the projection formula

    .. math::

        \psi_\theta(\sigma) = \frac{1}{|G|}\sum_{g\in G}
            \chi_g \psi_\theta(T_{g}\sigma).

    Ground states usually transform according to the trivial irrep
    :math:`\chi_g=1 \forall g\in G`.

    Examples:

        Symmetrise an :ref:`netket.models.RBM` with respect to the space
        group of a 2D :ref:`netket.graph.Square` lattice:

        >>> import netket as nk
        >>> graph = nk.graph.Square(4)
        >>> group = graph.space_group()
        >>> print("Size of space group:", len(group))
        Size of space group: 128
        >>> # Construct the bare unsymmetrized machine
        >>> machine_no_symm = nk.models.RBM(alpha=2)
        >>> # Symmetrize the RBM over the space group
        >>> ma = nk.nn.blocks.SymmExpSum(module = machine_no_symm, symm_group=group)

        Nontrivial irreps can be specified using momentum and point-group
        quantum numbers:

        >>> from math import pi
        >>> print(group.little_group(pi, 0).character_table_readable())
        (['1xId()', '1xRefl(0°)', '1xRefl(90°)', '1xRot(180°)'],
        array([[ 1.,  1.,  1.,  1.],
               [ 1.,  1., -1., -1.],
               [ 1., -1.,  1., -1.],
               [ 1., -1., -1.,  1.]]))
        >>> chi = group.space_group_irreps(pi, 0)[1]
        >>> ma = nk.nn.blocks.SymmExpSum(module = machine_no_symm, symm_group=group, characters=chi)

        Convolutional networks are already invariant under translations,
        so they only need to be symmetrised with respect to the point group
        (e.g., mirrors and rotations).

        >>> import netket as nk
        >>> graph = nk.graph.Square(4)
        >>> print("Size of the point group:", len(graph.point_group()))
        Size of the point group: 8
        >>> # Construct a translation-invariant RBM
        >>> machine_trans = nk.models.RBMSymm(alpha=2, symmetries=graph.translation_group())
        >>> # Symmetrize the RBM over the point group
        >>> ma = nk.nn.blocks.SymmExpSum(module = machine_trans, symm_group=graph.point_group())
    """

    module: nn.Module
    """The unsymmetrised neural-network ansatz."""

    symm_group: PermutationGroup
    """The symmetry group to use. It should be a valid
    :ref:`netket.utils.group.PermutationGroup` object.

    Can be extracted from a :ref:`netket.graph.Lattice` object by calling
    :meth:`~netket.graph.Lattice.point_group` or
    :meth:`~netket.graph.Lattice.translation_group`.

    Alternatively, if you have a :class:`netket.graph.Graph` object you
    can build it from :meth:`~netket.graph.Lattice.automorphisms`.

    .. code::

        graph = nk.graph.Square(3)
        symm_group = graph.point_group()

    """

    characters: HashableArray | None = None
    r"""Characters :math:`\chi_g` of the space group to project onto.

    Only one of `characters` and `character_id` may be specified.
    If neither is specified, the character is taken to be all 1,
    yielding a trivially symmetric state.
    """

    character_id: int | None = None
    """Index of the character to project onto in the character table
    of the symmetry group.

    The characters are accessed as:

    .. code::

        symm_group.character_table()[character_id]

    Only one of `characters` and `character_id` may be specified.
    If neither is specified, the character is taken to be all 1,
    yielding a trivially symmetric state.
    """

    def setup(self):
        if self.characters is None:
            if self.character_id is None:
                self._chi = np.ones(len(np.asarray(self.symm_group)))
            else:
                self._chi = self.symm_group.character_table()[self.character_id]
        else:
            if self.character_id is None:
                self._chi = self.characters.wrapped
            else:
                raise AttributeError(
                    "Must not specify both `characters` and `character_id`"
                )

    @nn.compact
    def __call__(self, x: jax.Array) -> jax.Array:
        """
        Accepts a single input or arbitrary batch of inputs.

        The last dimension of x must match the shape of the permutation
        group.
        """
        # apply the group and obtain a x_symm of shape (N_symm, ...batch_dims..., N_sites)
        x_symm = self.symm_group @ x

        # Apply the vmapped module to x_symm so that we don't 'see' this extra dim.
        psi_symm = jax.vmap(self.module)(x_symm)

        characters = np.expand_dims(self._chi, tuple(range(1, x.ndim)))

        # If those are all positive, then use standard logsumexp that returns a
        # real-valued, positive logsumexp
        logsumexp_fun = (
            jax.scipy.special.logsumexp if np.all(characters >= 0) else logsumexp_cplx
        )

        # log (sum_i ( c_i/Nsymm* exp(psi[sigma_i])))
        psi = logsumexp_fun(psi_symm, axis=0, b=characters / len(self.symm_group))
        return psi



from collections.abc import Callable
from functools import partial

import jax
import jax.numpy as jnp
from jax.sharding import NamedSharding, PartitionSpec as P

from netket import jax as nkjax
from netket import config
from netket.utils import timing
from netket.utils.types import Array


@timing.timed
@partial(
    jax.jit,
    static_argnames=(
        "solver_fn",
        "mode",
    ),
)
def _compute_srt_update(
    O_L,
    dv,
    *,
    diag_shift: float | Array,
    solver_fn: Callable[[Array, Array], Array],
    mode: str,
    proj_reg: float | Array | None = None,
    momentum: float | Array | None = None,
    old_updates: Array | None = None,
    params_structure,
):
    if momentum is not None:
        dv -= momentum * (O_L @ old_updates)

    # Here we reshard the jacobian from being sharded along the sample axis to being
    # sharded along the parameter axis, which we pad
    # (#ns, np) -> (ns, #np)
    # In theory jax could figure this out, but maybe he does not, so we do it by hand.
    O_LT = O_L
    if config.netket_experimental_sharding:
        O_LT = nkjax.sharding.pad_axis_for_sharding(O_LT, axis=1, padding_value=0.0)
        O_LT = jax.lax.with_sharding_constraint(
            O_LT,
            NamedSharding(jax.sharding.get_abstract_mesh(), P(None, "S")),
        )
        dv = jax.lax.with_sharding_constraint(
            dv, NamedSharding(jax.sharding.get_abstract_mesh(), P())
        )

    # This does the contraction (ns, #np) x (#np, ns) -> (ns, ns).
    # When using sharding the sum over #ns is done automatically.
    # When using MPI we need to do it manually with an allreduce_sum.
    matrix = O_LT @ O_LT.T
    matrix_side = matrix.shape[-1]  # * it can be ns or 2*ns, depending on mode

    shifted_matrix = jax.lax.add(
        matrix, diag_shift * jnp.eye(matrix_side, dtype=matrix.dtype)
    )
    # replicate

    if proj_reg is not None:
        shifted_matrix = jax.lax.add(
            shifted_matrix, jnp.full_like(shifted_matrix, proj_reg / matrix_side)
        )

    aus_vector = solver_fn(shifted_matrix, dv)

    # Some solvers return a tuple, some others do not.
    if isinstance(aus_vector, tuple):
        aus_vector, info = aus_vector
        if info is None:
            info = {}
    else:
        info = {}

    # (np, #ns) x (#ns) -> (np).
    updates = O_L.T @ aus_vector
    if momentum is not None:
        updates += momentum * old_updates
        old_updates = updates

    # If complex mode and we have complex parameters, we need
    # To repack the real coefficients in order to get complex updates
    if mode == "complex" and nkjax.tree_leaf_iscomplex(params_structure):
        num_p = updates.shape[-1] // 2
        updates = updates[:num_p] + 1j * updates[num_p:]

    if config.netket_experimental_sharding:
        out_shardings = NamedSharding(
            jax.sharding.get_abstract_mesh(), P(*(None,) * updates.ndim)
        )
        updates = jax.lax.with_sharding_constraint(updates, out_shardings)

    return updates, old_updates, info



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Ignore false-positives for redefined `product` functions:
# pylint: disable=function-redefined

import numpy as np
from functools import reduce
from collections.abc import Iterable, Sequence
from warnings import warn

from netket.utils import struct, deprecated_new_name, deprecated
from netket.utils.types import Array
from netket.utils.float import prune_zeros
from netket.utils.dispatch import dispatch
from netket.utils.group import (
    Element,
    Identity,
    PointGroup,
    Permutation,
    PermutationGroup,
)

from .lattice import Lattice


class Translation(Permutation):
    r"""
    Custom subclass of `Permutation` that represents a lattice permutation.
    Stores translation lattice vector and generates a sensible name from it.

    The product of two `Translation`s carries the appropriate displacement vector.
    """

    def __init__(
        self,
        permutation: Array | None = None,
        *,  # maybe change something
        displacement: Array,
        permutation_array: Array | None = None,
        inverse_permutation_array: Array | None = None,
    ):
        r"""
        Creates a `Translation` from either the array of images
        `permutation_array` or preimages `inverse_permutation_array`,
        and a displacement vector.

        Exactly one argument among `permutation_array` and
        `inverse_permutation_array` (and the deprecated argument `permutation`)
        must be specified.

        The deprecated argument `permutation` should be substituted for
        `inverse_permutation_array`.

        Note that the left action of a permutation on an array `a` is
        `a[inverse_permutation_array]`.

        Args:
            permutation: (deprecated) 1D array listing
                :math:`g^{-1}(x)` for all :math:`0\le x \le N-1`.
            displacement: displacement vector is units of lattice basis vectors
            permutation_array: 1D array listing
                :math:`g(x)` for all :math:`0\le x \le N-1`.
            inverse_permutation_array: 1D array listing
                :math:`g^{-1}(x)` for all :math:`0\le x \le N-1`.

        Returns:
            A `Translation` object that encodes the specified translation.
        """
        super().__init__(
            permutation_array=permutation_array,
            inverse_permutation_array=inverse_permutation_array,
            permutation=permutation,
        )
        self._vector = np.asarray(displacement)

    @property
    def _name(self):
        return f"Translation({self._vector.tolist()})"


@dispatch
def product(p: Translation, q: Translation):
    inverse_permutation_array = q.inverse_permutation_array[p.inverse_permutation_array]
    return Translation(
        inverse_permutation_array=inverse_permutation_array,
        displacement=p._vector + q._vector,
    )


def _ensure_iterable(x):
    """Extracts iterables given in varargs"""
    if isinstance(x[0], Iterable):
        if len(x) > 1:
            raise TypeError("Either Iterable or variable argument list expected")
        return x[0]
    else:
        return x


# This function doesn't seem to be tested
def _translations_along_axis(lattice: Lattice, axis: int) -> PermutationGroup:
    """
    The group of valid translations along an axis as a `PermutationGroup`
    acting on the sites of `lattice.`
    """
    if lattice._pbc[axis]:
        trans_list = [Identity()]
        # note that we need the preimages in the permutation
        trans_perm = lattice.id_from_position(
            lattice.positions - lattice.basis_vectors[axis]
        )
        vector = np.zeros(lattice.ndim, dtype=int)
        vector[axis] = 1
        trans_by_one = Translation(
            inverse_permutation_array=trans_perm, displacement=vector
        )

        for _ in range(1, lattice.extent[axis]):
            trans_list.append(trans_list[-1] @ trans_by_one)

        return PermutationGroup(trans_list, degree=lattice.n_nodes)
    else:
        return PermutationGroup([Identity()], degree=lattice.n_nodes)


# This function doesn't seem to be tested
def _pg_to_permutation(lattice: Lattice, point_group: PointGroup) -> PermutationGroup:
    """
    The permutation action of `point_group` on the sites of `lattice`.
    """
    perms: list[Element] = []
    for p in point_group:
        if isinstance(p, Identity):
            perms.append(Identity())
        else:
            # note that we need the preimages in the permutation
            perm = lattice.id_from_position(p.preimage(lattice.positions))
            perms.append(
                Permutation(inverse_permutation_array=perm, name=str(p), validate=False)
            )
    return PermutationGroup(perms, degree=lattice.n_nodes)


@struct.dataclass
class TranslationGroup(PermutationGroup):
    """
    Class to handle translation symmetries of a :class:`~netket.graph.Lattice`.

    Corresponds to a representation of the translation group
    on the given lattice as a permutation group of :code:`N_sites` variables.

    Can be used as a :class:`~netket.utils.group.PermutationGroup` representing
    the translations, but the product table is computed much more efficiently
    than in a generic :class:`~netket.utils.group.PermutationGroup`.
    """

    lattice: Lattice
    """The lattice whose translation group is represented."""
    axes: tuple[int]
    """Axes translations along which are represented by the group."""

    def __repr__(self):
        return type(self).__name__ + f"(lattice:\n{self.lattice}\naxes:{self.axes})"

    def __pre_init__(
        self, lattice: Lattice, axes: int | tuple[int] | None = None
    ) -> tuple[tuple, dict]:
        if axes is None:
            axes = tuple(range(lattice.ndim))
        elif isinstance(axes, int):
            axes = [axes]

        if len(set(axes)) != len(axes):
            raise ValueError(
                f"Axes must be unique integers, they cannot repeat (got `{axes}`)."
            )
        if not all(0 <= x < lattice.ndim for x in axes):
            raise ValueError(
                f"Axes must be integers in the range [0, {lattice.ndim}) (got `{axes}`)."
            )

        # compute translation group by axis and overall
        translation_by_axis = [_translations_along_axis(lattice, i) for i in axes]
        translation_group = reduce(PermutationGroup.__matmul__, translation_by_axis)

        return (), dict(
            lattice=lattice,
            axes=tuple(axes),
            elems=translation_group.elems,
            degree=lattice.n_nodes,
        )

    def __hash__(self):
        return super().__hash__()

    @struct.property_cached
    def group_shape(self) -> Array:
        """
        Tuple of the number of translations represented by the group along
        each lattice direction.

        :code:`self.group_shape[i]` is :code:`self.lattice.extent[i]`
        if both :code:`i in self.axes` and :code:`self.lattice.pbc[i] is True`,
        otherwise 1.
        """
        axes_bool = np.zeros(self.lattice.ndim, dtype=bool)
        axes_bool[list(self.axes)] = True
        in_group = np.logical_and(self.lattice.pbc, axes_bool)
        shape = np.where(in_group, self.lattice.extent, 1)
        return shape

    @struct.property_cached
    def inverse(self) -> Array:
        ix = np.ix_(*[-np.arange(x) % x for x in self.group_shape])
        inv = np.arange(len(self)).reshape(self.group_shape)[ix]
        return np.ravel(inv)

    @struct.property_cached
    def product_table(self) -> Array:
        # product table of each 1d component
        ix = [(np.arange(x) - np.arange(x)[:, None]) % x for x in self.group_shape]
        ix = np.ix_(*[np.ravel(x) for x in ix])
        # product table with n_axes axes
        # each axis stands for row and column direction along one axis
        pt = np.arange(len(self)).reshape(self.group_shape)[ix]
        shape = [x for x in self.group_shape for _ in range(2)]
        # separate row and column directions
        pt = pt.reshape(shape)
        # bring all row and all column directions together
        pt = pt.transpose(list(range(0, len(shape), 2)) + list(range(1, len(shape), 2)))

        return pt.reshape(len(self), len(self))

    def momentum_irrep(self, *k: Array) -> np.ndarray:
        r"""Returns the irrep characters (phase factors) corresponding to
        crystal momentum :math:`\vec k`."""
        # switch to reciprocal lattice coordinates
        k = self.lattice.to_reciprocal_lattice(_ensure_iterable(k)).squeeze()

        # prune axes with no nontrivial translations for performance
        shape = np.asarray(self.group_shape)
        nontrivial_axes = shape > 1
        k = k[nontrivial_axes]
        shape = shape[nontrivial_axes]

        # phase factors for translations along each axis
        axis_factors = [
            np.exp(-2j * np.pi * ki * np.arange(ni) / ni) for ki, ni in zip(k, shape)
        ]
        axis_factors = np.ix_(*axis_factors)

        return reduce(np.multiply, axis_factors).ravel()


_tg_efficiency_notice = """

        Computed more efficiently than for a generic
        :class:`~netket.utils.group.PermutationGroup` exploiting the
        Abelian group structure."""
TranslationGroup.inverse.__doc__ = (
    PermutationGroup.inverse.__doc__ + _tg_efficiency_notice
)
TranslationGroup.product_table.__doc__ = (
    PermutationGroup.product_table.__doc__ + _tg_efficiency_notice
)


@struct.dataclass
class SpaceGroup(PermutationGroup):
    """
    Class to handle the space group symmetries of a :class:`~netket.graph.Lattice`.

    Can be used as a :class:`~netket.utils.group.PermutationGroup`
    representing the action of a space group on a :class:`~netket.graph.Lattice`.
    The space group is generated as the semidirect product of the translation group
    of the lattice and a geometrical :class:`~netket.utils.group.PointGroup`
    given as a constructor argument.

    Also generates :class:`~netket.utils.group.PermutationGroup` representations of

    * the supplied point group,
    * its rotational subgroup (i.e. point group symmetries with determinant +1)
    * the translation group of the lattice

    as well as space group irreps for symmetrising wave functions.
    """

    lattice: Lattice
    """The lattice underlying the space group."""
    _point_group: PointGroup
    """The geometric point group underlying the space group."""
    point_group: PermutationGroup
    """The point group as a :class:`~netket.utils.group.PermutationGroup`
    acting on the sites of :attr:`lattice`.

    Group elements are listed in the order they appear in :attr:`_point_group`.
    Computed from :attr:`_point_group` upon construction, must not be changed after."""
    full_translation_group: TranslationGroup

    def __pre_init__(
        self, lattice: Lattice, point_group: PointGroup
    ) -> tuple[tuple, dict]:
        """
        Constructs the space group generated by the translation symmetries of
        the lattice and a given point group.

        Args:
            lattice: :class:`~netket.graph.Lattice`
                The lattice on which the space group is to act.
            point_group: :class:`~netket.utils.group.PointGroup`
                The geometrical point group underlying the space group.
        """

        if not isinstance(lattice, Lattice):
            raise TypeError(
                "Expected an instance of `Lattice` as argument `lattice`, "
                f"got {type(lattice)}."
            )

        if not isinstance(point_group, PointGroup):
            msg = (
                "Expected an instance of `PointGroup` as argument `point_group`, "
                f"got type {type(point_group)}."
            )
            if isinstance(point_group, Element):
                msg += (
                    "\n\n`point_group` appears to be a single symmetry. "
                    "It should be wrapped into a `PointGroup` object."
                )
            raise TypeError(msg)

        # compute point group permutations
        point_group = point_group.replace(unit_cell=lattice.basis_vectors)
        pg_as_perm = _pg_to_permutation(lattice, point_group)

        # compute translation group by axis and overall
        translation_group = lattice.full_translation_group

        # compute space group elements
        space_group = translation_group @ pg_as_perm
        elems = space_group.elems

        return (), dict(
            lattice=lattice,
            _point_group=point_group,
            point_group=pg_as_perm,
            full_translation_group=translation_group,
            elems=elems,
            degree=lattice.n_nodes,
        )

    def __hash__(self):
        return super().__hash__()

    def __repr__(self):
        return (
            type(self).__name__
            + f"(lattice:\n{self.lattice}\npoint_group:\n{self._point_group})"
        )

    @property
    @deprecated_new_name("_point_group", reason="Consistency")
    def point_group_(self) -> PointGroup:
        """
        Deprecated: Returns the internally stored point group as a point group,
        instead of the one stored as a permutation group.
        """

        return self._point_group

    # TODO describe ordering of group elements here and later in docstring

    @struct.property_cached
    def rotation_group(self) -> PermutationGroup:
        """The group of rotations (i.e. point group symmetries with determinant +1)
        as a :class:`~netket.utils.group.PermutationGroup` acting on the sites of :attr:`lattice`.

        Group elements are listed in the order they appear in :attr:`_point_group`."""
        return _pg_to_permutation(self.lattice, self._point_group.rotation_group())

    def translation_group(
        self, axes: int | Sequence[int] | None = None
    ) -> TranslationGroup:
        """
        The group of valid translations of :attr:`lattice` as a :class:`TranslationGroup`
        acting on the sites of the same.
        """
        if axes is None:
            return self.full_translation_group
        else:
            return TranslationGroup(self.lattice, axes=axes)

    @deprecated_new_name("translation_group")
    def _translations_along_axis(self, axis: int) -> TranslationGroup:
        # DEPRECATED: use `self.translation_group(axes=axis)` instead
        return self.translation_group(axes=axis)

    @property
    @deprecated(
        reason="This `SpaceGroup` object can be used directly as a permutation group"
    )
    def space_group(self) -> "SpaceGroup":
        """
        Deprecated. Returns :code:`self`.
        """
        return self

    @struct.property_cached
    def product_table(self) -> Array:
        # compute first n_PG rows of product table like in PermutationGroup
        perms = self.to_array()
        inverse = perms[self.inverse].squeeze()
        n_symm = len(perms)
        n_PG = len(self._point_group)
        n_TG = len(self.full_translation_group)
        lookup = np.unique(np.column_stack((perms, np.arange(len(self)))), axis=0)

        PG_rows = np.zeros([n_PG, n_symm], dtype=int)
        for i, g_inv in enumerate(inverse[:n_PG]):
            row_perms = perms[:, g_inv]
            row_perms = np.unique(
                np.column_stack((row_perms, np.arange(len(self)))), axis=0
            )
            # row_perms should be a permutation of perms, so identical after sorting
            if np.any(row_perms[:, :-1] != lookup[:, :-1]):
                raise RuntimeError(
                    "PermutationGroup is not closed under multiplication"
                )
            # match elements in row_perms to group indices
            PG_rows[i, row_perms[:, -1]] = lookup[:, -1]

        # PG_rows contains pg^-1 th ph - split three terms into three dimensions
        PG_rows = PG_rows.reshape(n_PG, n_TG, n_PG)
        # the full product table is of the form pg^-1 tg^-1 th ph
        # the middle two terms are the product table of the TG
        product_table = PG_rows[:, self.full_translation_group.product_table, :]
        # reshuffle into output shape
        product_table = product_table.transpose(1, 0, 2, 3)

        return product_table.reshape(n_symm, n_symm)

    @struct.property_cached
    def _point_group_conjugacy_table(self) -> np.ndarray:
        """Part of the conjugacy table :math:`h^{-1}gh` where h are
        point-group symmetries."""
        n_PG = len(self.point_group)
        col_index = np.arange(n_PG)[np.newaxis, :]
        # exploits that h^{-1}gh = (g^{-1} h)^{-1} h
        return self.product_table[self.product_table[:, :n_PG], col_index]

    def _little_group_index(self, k: Array) -> Array:
        r"""
        Returns the indices of the elements of the little group corresponding to
        wave vector :math:`\vec{k}`.
        """
        # calculate k' = p(k) for all p in the point group
        big_star = np.tensordot(self._point_group.matrices(), k, axes=1)
        big_star = self.lattice.to_reciprocal_lattice(big_star) % self.lattice.extent
        # should test for pbc before taking the modulus, but the only valid wave
        # vector for non-pbc axes is 0 and 0 % anything == 0

        # assumes point_group_[0] is the identity
        is_in_little_group = np.all(big_star == big_star[0], axis=1)
        return np.arange(len(self._point_group))[is_in_little_group]

    def little_group(self, *k: Array) -> PointGroup:
        r"""
        Returns the little co-group corresponding to wave vector :math:`\vec{k}`.
        This is the subgroup of the point group that leaves :math:`\vec{k}` invariant.

        Arguments:
            k: the wave vector in Cartesian axes

        Returns:
            the little co-group
        """
        k = _ensure_iterable(k)
        return PointGroup(
            [self._point_group[i] for i in self._little_group_index(k)],
            ndim=self._point_group.ndim,
            unit_cell=self.lattice.basis_vectors,
        )

    def little_group_multipliers(self, *k: Array) -> np.ndarray | None:
        r"""Computes the Schur multiplier associated with the little group
        given the translations associated with its elements.

        The mutlipliers are given by (Bradney & Cracknell, eqs. 3.7.11-14)

        .. math::

            \mu(S_i, S_j) &= \exp(-i g_i \cdot w_j)

            g_i  &= S_i^{-1} k - k

        and :math:`w_j` is the translation associated with
        point-group symmetry :math:`S_i`.

        Arguments:
            k: the wave vector in Cartesian axes

        Returns:
            A square array of the :math:`\mu(S_i, S_j)`.

            If all multipliers are +1, :code:`None` is returned instead:
            this signals to :meth:`~netket.utils.group.FiniteGroup.character_table()`
            etc. that linear (not projective) representations are required.
        """
        k = _ensure_iterable(k)
        ix = self._little_group_index(k)

        # g_i  = S_i^{-1} k - k
        matrices = self._point_group.matrices()[ix]
        matrices = matrices.transpose(0, 2, 1)  # need S_i^{-1}
        g = matrices @ k - k

        w = self._point_group.translations()[ix]

        multiplier = np.exp(-1j * (g @ w.T))

        if np.allclose(multiplier, 1.0, rtol=1e-8):
            return None
        else:
            return prune_zeros(multiplier)

    def little_group_irreps_readable(self, *k: Array, full: bool = False):
        """Returns a conventional rendering of little-group irrep characters.

        This differs from :code:`little_group(k).character_table_readable()`
        in that nontrivial Schur multipliers for nonsymmorphic space group
        are automatically taken into account.

        Arguments:
            k: the wave vector in Cartesian axes
            full: whether the character table for all group elements (True)
                or one representative per conjugacy class (False, default)

        Returns:

            A tuple containing a list of strings and an array

            - :code:`classes`: a text description of a representative of
              each conjugacy class (or each element) of the little group as a list
            - :code:`characters`: a matrix, each row of which lists the
              characters of one irrep
        """
        k = _ensure_iterable(k)
        group = self.little_group(k)
        multiplier = self.little_group_multipliers(k)

        if multiplier is not None:
            warn(
                "The space group is nonsymmorphic and the function will return\n"
                "a character table of projective irreps of the little group.\n"
                "If you want the linear irreps of the little group, use\n"
                "`self.little_group(k).character_table_readable()` instead."
            )

        return group.character_table_readable(multiplier, full)

    def _little_group_irreps(self, k: Array) -> Array:
        """
        Returns the character table of the little group embedded in the full point
        group. Symmetries outside the little group get 0.
        """
        idx = self._little_group_index(k)
        group = self.little_group(k)
        multiplier = self.little_group_multipliers(k)
        CT = group.character_table(multiplier)
        CT_full = np.zeros((CT.shape[0], len(self._point_group)), dtype=CT.dtype)
        CT_full[:, idx] = CT
        return CT_full

    def space_group_irreps(self, *k: Array) -> Array:
        r"""
        Returns the portion of the character table of the full space group
        corresponding to the star of the wave vector :math:`\vec{k}`.

        Arguments:
            k: the wave vector in Cartesian axes

        Returns:
            An array :code:`CT` listing the characters for all irreps of the
            space group defined on the star of :math:`\vec{k}`.

            :code:`CT[i]` returns the irrep corresponding to the little-group
            irrep listed in row #i by :meth:`little_group_irreps_readable`.

            :code:`CT[i,j]` gives the character of :code:`self[j]` in the same.
        """
        # One-arm irreps for the other wave vectors in the star can be
        # obtained from the arm `k` by conjugating the character with some
        # point-group operation.
        # The simplest is to do the conjugation with all of them, which
        # counts every arm |little group| times, which we divide out
        # at the end.
        k = _ensure_iterable(k)
        chi = self.one_arm_irreps(k)[:, self._point_group_conjugacy_table]
        return prune_zeros(chi.sum(axis=-1) / len(self._little_group_index(k)))

    def one_arm_irreps(self, *k: Array) -> Array:
        r"""
        Returns the portion of the character table of the full space group
        corresponding to the star of the wave vector :math:`\vec{k}`,
        projected onto :math:`\vec{k}` itself.

        Arguments:
            k: the wave vector in Cartesian axes

        Returns:
            An array `CT` listing the projected characters for all irreps of the
            space group defined on the star of :math:`\vec{k}`.

            :code:`CT[i]` returns the irrep corresponding to the little-group
            irrep listed in row #i by :meth:`little_group_irreps_readable`.

            :code:`CT[i,j]` gives the character of :code:`self[j]` in the same.
        """
        k = _ensure_iterable(k)
        # Little-group irrep factors
        # Phase factor for non-symmorphic symmetries is exp(-i w_g . p(k))
        point_group_factors = self._little_group_irreps(k) * np.exp(
            -1j * (self._point_group.translations() @ k)
        )
        # Translational factors
        trans_factors = self.full_translation_group.momentum_irrep(k)

        # Multiply the factors together
        # Translations are more major than point group operations
        result = np.einsum("ig, t -> itg", point_group_factors, trans_factors)
        result = result.reshape(point_group_factors.shape[0], -1)
        return prune_zeros(result)


_sg_efficiency_notice = """

        Computed more efficiently than for a generic
        :class:`~netket.utils.group.PermutationGroup` exploiting the
        semidirect product structure of space groups."""
SpaceGroup.product_table.__doc__ = (
    PermutationGroup.product_table.__doc__ + _sg_efficiency_notice
)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Cubic and related 3D symmetry groups.

This module provides functions to construct symmetry groups for cubic systems,
including tetrahedral, octahedral, and icosahedral point groups. It includes
both pure rotational groups and full point groups with inversions, as well as
space groups for crystal lattices like diamond and pyrochlore.
"""

import numpy as np

from netket.utils.moduletools import export, hide_unexported

from netket.utils.group.axial import cuboid, cuboid_rotations  # noqa: F401
from netket.utils.group.axial import inversion_group as _inv_group
from netket.utils.group.axial import reflection_group as _refl_group
from netket.utils.group.axial import rotation as _rotation
from netket.utils.group._point_group import PointGroup
from netket.utils.group._semigroup import Identity


hide_unexported(__name__)

__all__ = [
    "tetrahedral_rotations",
    "tetrahedral",
    "pyritohedral",
    "octahedral_rotations",
    "cubic_rotations",
    "octahedral",
    "cubic",
    "diamond",
    "pyrochlore",
    "Fd3m",
]


@export
def T() -> PointGroup:
    """
    Rotational symmetries of a tetrahedron with vertices
    (1,1,1), (1,-1,-1), (-1,1,-1), (-1,-1,1).
    """
    return PointGroup(
        [
            Identity(),
            _rotation(120, [1, 1, 1]),
            _rotation(120, [1, -1, -1]),
            _rotation(120, [-1, 1, -1]),
            _rotation(120, [-1, -1, 1]),
            _rotation(-120, [1, 1, 1]),
            _rotation(-120, [1, -1, -1]),
            _rotation(-120, [-1, 1, -1]),
            _rotation(-120, [-1, -1, 1]),
            _rotation(180, [0, 0, 1]),
            _rotation(180, [0, 1, 0]),
            _rotation(180, [1, 0, 0]),
        ],
        ndim=3,
    )


tetrahedral_rotations = T


@export
def Td() -> PointGroup:
    r"""
    Symmetry group of a tetrahedron with vertices
    (1,1,1), (1,-1,-1), (-1,1,-1), (-1,-1,1).
    """
    return _refl_group([1, 1, 0]) @ T()


tetrahedral = Td


@export
def Th() -> PointGroup:
    """Pyritohedral symmetry group generated by T and inversion."""
    return _inv_group() @ T()


pyritohedral = Th


@export
def O() -> PointGroup:  # noqa: E741, E743
    """Rotational symmetries of a cube/octahedron aligned with the Cartesian axes."""
    # NB the first factor isn't an actual point group but this is fine
    # we only use it to generate a coset of T in O
    return PointGroup([Identity(), _rotation(90, [0, 0, 1])], ndim=3) @ T()


octahedral_rotations = cubic_rotations = O


@export
def Oh() -> PointGroup:
    """Symmetry group of a cube/octahedron aligned with the Cartesian axes."""
    return _inv_group() @ O()


octahedral = cubic = Oh


@export
def Fd3m() -> PointGroup:
    """Nonsymmorphic "point group" of the diamond and pyrochlore lattices with a
    cubic unit cell of side length 1."""
    return (_inv_group().change_origin([1 / 8, 1 / 8, 1 / 8]) @ Td()).replace(
        unit_cell=np.asarray([[0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]])
    )


diamond = pyrochlore = Fd3m



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import jax

from netket.optimizer.linear_operator import DenseOperator
from netket.utils.api_utils import partial_from_kwargs

from netket._src.vqs.fermion_mf.state import DeterminantVariationalState


@partial_from_kwargs
def QGTByWick(
    vstate: DeterminantVariationalState,
    *,
    diag_shift: float = 0.0,
    **kwargs,
) -> DenseOperator:
    """
    Computes the Quantum Geometric Tensor (QGT) for a fermionic mean-field state
    analytically using Wick's theorem.

    This function follows the same API pattern as other QGT constructors like
    :func:`~netket.optimizer.qgt.QGTJacobianDense`.

    Args:
        vstate: The DeterminantVariationalState for which to compute the QGT.
        diag_shift: Diagonal shift to add to the QGT matrix for regularization.
            This is equivalent to adding a small positive value to the diagonal
            to ensure the matrix is positive definite (default: 0.0).
        **kwargs: Additional keyword arguments (currently unused, for API compatibility).

    Returns:
        A DenseOperator wrapping the dense QGT matrix.

    """
    if kwargs.pop("diag_scale", None) is not None:
        raise NotImplementedError(
            "\n`diag_scale` argument is not yet supported by QGTByWick.\n"
        )

    qgt_matrix = _jit_apply(
        vstate._model,
        vstate.variables,
        method=vstate._model.qgt_by_wick,
    )

    return DenseOperator(matrix=qgt_matrix, diag_shift=diag_shift)


@partial(jax.jit, static_argnames=("model", "method"))
def _jit_apply(model, *args, method):
    return model.apply(*args, method=method)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Sequence
from typing import TypeVar as _TypeVar

import numpy as np

from netket.graph import AbstractGraph, Graph
from netket.hilbert import AbstractHilbert
from netket.utils.types import DType

from ._graph_operator import GraphOperator
from ._local_operator import (
    LocalOperator as _LocalOperator,
    LocalOperatorBase as _LocalOperatorBase,
)

_LocalOperatorT = _TypeVar("_LocalOperatorT", bound=_LocalOperatorBase)


def Heisenberg(
    hilbert: AbstractHilbert,
    graph: AbstractGraph,
    J: float | Sequence[float] = 1.0,
    sign_rule: None | bool | Sequence[bool] = None,
    dtype: DType | None = None,
    *,
    acting_on_subspace: None | list[int] | int = None,
    cls: type[_LocalOperatorT] = _LocalOperator,
) -> _LocalOperatorT:
    """
    Constructs an Heisenberg operator given a hilbert space and a graph providing the
    connectivity of the lattice.

    Args:
        hilbert: Hilbert space the operator acts on.
        graph: The graph upon which this hamiltonian is defined.
        J: The strength of the coupling. Default is 1.
           Can pass a sequence of coupling strengths with coloured graphs:
           edges of colour n will have coupling strength J[n]
        sign_rule: If True, Marshal's sign rule will be used. On a bipartite
            lattice, this corresponds to a basis change flipping the Sz direction
            at every odd site of the lattice. For non-bipartite lattices, the
            sign rule cannot be applied. Defaults to True if the lattice is
            bipartite, False otherwise.
            If a sequence of coupling strengths is passed, defaults to False
            and a matching sequence of sign_rule must be specified to override it
        dtype: Data type of the matrix elements.
        acting_on_subspace: Specifies the mapping between nodes of the graph and
            Hilbert space sites, so that graph node :code:`i ∈ [0, ..., graph.n_nodes - 1]`,
            corresponds to :code:`acting_on_subspace[i] ∈ [0, ..., hilbert.n_sites]`.
            Must be a list of length `graph.n_nodes`. Passing a single integer :code:`start`
            is equivalent to :code:`[start, ..., start + graph.n_nodes - 1]`.

    Examples:
     Constructs a ``Heisenberg`` operator for a 1D system.

        >>> import netket as nk
        >>> g = nk.graph.Hypercube(length=20, n_dim=1, pbc=True)
        >>> hi = nk.hilbert.Spin(s=0.5, total_sz=0, N=g.n_nodes)
        >>> op = nk.operator.Heisenberg(hilbert=hi, graph=g)
        >>> print(op)
        LocalOperatorJax(dim=20, #acting_on=20 locations, constant=0.0, dtype=float64)
    """
    if isinstance(J, Sequence):
        # check that the number of Js matches the number of colours
        assert len(J) == max(graph.edge_colors) + 1

        if sign_rule is None:
            sign_rule = [False] * len(J)
        else:
            assert len(sign_rule) == len(J)
            for i in range(len(J)):
                subgraph = Graph(edges=graph.edges(filter_color=i))
                if sign_rule[i] and not subgraph.is_bipartite():
                    raise ValueError(
                        "sign_rule=True specified for a non-bipartite lattice"
                    )
    else:
        if sign_rule is None:
            sign_rule = graph.is_bipartite()
        elif sign_rule and not graph.is_bipartite():
            raise ValueError("sign_rule=True specified for a non-bipartite lattice")

    sz_sz = np.array(
        [
            [1, 0, 0, 0],
            [0, -1, 0, 0],
            [0, 0, -1, 0],
            [0, 0, 0, 1],
        ]
    )
    exchange = np.array(
        [
            [0, 0, 0, 0],
            [0, 0, 2, 0],
            [0, 2, 0, 0],
            [0, 0, 0, 0],
        ]
    )

    if isinstance(J, Sequence):
        bond_ops = [
            J[i] * (sz_sz - exchange if sign_rule[i] else sz_sz + exchange)
            for i in range(len(J))
        ]
        bond_ops_colors = list(range(len(J)))
    else:
        bond_ops = [J * (sz_sz - exchange if sign_rule else sz_sz + exchange)]
        bond_ops_colors = []

    return GraphOperator(
        hilbert,
        graph,
        bond_ops=bond_ops,
        bond_ops_colors=bond_ops_colors,
        dtype=dtype,
        acting_on_subspace=acting_on_subspace,
        cls=cls,
    )



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Compute empirical NNGP and NTK; approximate functions via Taylor series.

All functions in this module are applicable to any JAX functions of proper
signatures (not only those from :obj:`~neural_tangents.stax`).

NNGP and NTK are computed using :obj:`~neural_tangents.empirical_nngp_fn`,
:obj:`~neural_tangents.empirical_ntk_fn`, or
:obj:`~neural_tangents.empirical_kernel_fn` (for both). The kernels have a very
specific output shape convention that may be unexpected. Further, NTK has
multiple implementations that may perform differently depending on the task.
Please read individual functions' docstrings.

For details, please see "`Fast Finite Width Neural Tangent Kernel
<https://arxiv.org/abs/2206.08720>`_".
"""

from typing import Any, Protocol, TypeVar, Union
from collections.abc import Callable, Iterable

from functools import partial
import operator

import jax
import jax.numpy as jnp
from jax.core import ShapedArray
from jax.interpreters.ad import UndefinedPrimal

from . import utils

safe_zip = partial(zip, strict=True)

PyTree = Any
_VMapAxis = PyTree | None
VMapAxisTriple = tuple[_VMapAxis, _VMapAxis, dict[str, _VMapAxis]]
VMapAxes = Union[_VMapAxis, VMapAxisTriple]
AnalyticKernelFn = Any
EmpiricalGetKernelFn = Any
MonteCarloKernelFn = Any
InitFn = Any
LayerKernelFn = Any
MaskFn = Any
Kernel = Any
Axes = Any
_ArrayOrShape = TypeVar("_ArrayOrShape", jnp.ndarray, ShapedArray)
T = TypeVar("T")
NTTree = Union[T, list["NTTree[T]"], tuple["NTTree[T]", ...], T]
NTTrees = Union[list["NTTree[T]"], tuple["NTTree[T]", ...]]


class ApplyFn(Protocol):
    """A type alias for apply functions.

    Apply functions do computations with finite-width neural networks. They are
    functions that take a PyTree of parameters and an array of inputs and produce
    an array of outputs.
    """

    def __call__(
        self, params: PyTree, inputs: NTTree[jnp.ndarray], *args, **kwargs
    ) -> NTTree[jnp.ndarray]: ...
class EmpiricalKernelFn(Protocol):
    """A type alias for empirical kernel functions computing either NTK or NNGP.

    A kernel function that produces an empirical kernel from a single
    instantiation of a neural network specified by its parameters.

    Equivalent to `EmpiricalGetKernelFn` with `get="nngp"` or `get="ntk"`.
    """

    def __call__(
        self,
        x1: NTTree[jnp.ndarray],
        x2: NTTree[jnp.ndarray] | None,
        params: PyTree,
        **kwargs,
    ) -> NTTree[jnp.ndarray]: ...


KernelFn = Union[
    AnalyticKernelFn,
    EmpiricalKernelFn,
    EmpiricalGetKernelFn,
    MonteCarloKernelFn,
]
InternalLayer = Any
InternalLayerMasked = tuple[InitFn, ApplyFn, LayerKernelFn, MaskFn]
Layer = tuple[InitFn, ApplyFn, AnalyticKernelFn]
Kernels = Union[list[Kernel], tuple[Kernel, ...]]
"Kernel inputs/outputs of `FanOut`, `FanInSum`, etc."
_VMapAxis = PyTree | None
"A `PyTree` of integers"
VMapAxisTriple = tuple[_VMapAxis, _VMapAxis, dict[str, _VMapAxis]]
VMapAxes = Union[_VMapAxis, VMapAxisTriple]
"Specifies `(input, output, kwargs)` axes for `vmap` in empirical NTK."


def empirical_ntk_by_jacobian(
    f: ApplyFn,
    trace_axes: Axes = (-1,),
    diagonal_axes: Axes = (),
    vmap_axes: VMapAxes = None,
    **kwargs,
) -> EmpiricalKernelFn:
    r"""Returns a function to draw a single sample the NTK of a given network `f`.

    The Neural Tangent Kernel is defined as :math:`J(X_1) J(X_2)^T` where
    :math:`J` is the Jacobian :math:`df/dparams` of shape
    `full_output_shape + params.shape`.

    Compute NTK by directly instantiating Jacobians and contracting.

    For best performance:
    1) pass `x2=None` if `x1 == x2;
    2) prefer square batches (i.e `x1.shape == x2.shape`);
    3) make sure to set `vmap_axes` correctly.

    .. warning::
      Resulting kernel shape is *nearly* `zip(f(x1).shape, f(x2).shape)`
      subject to `trace_axes` and `diagonal_axes` parameters, which make certain
      assumptions about the outputs `f(x)` that may only be true in the infinite
      width / infinite number of samples limit, or may not apply to your
      architecture. For most precise results in the context of linearized training
      dynamics of a specific finite-width network, set both `trace_axes=()` and
      `diagonal_axes=()` to obtain the kernel exactly of shape
      `zip(f(x1).shape, f(x2).shape)`.

    For networks with multiple (i.e. lists, tuples, PyTrees) outputs, in principal
    the empirical kernels will have terms measuring the covariance between the
    outputs. Here, we ignore these cross-terms and consider each output
    separately. Please raise an issue if this feature is important to you.

    Args:
      f:
        the function whose NTK we are computing. It should have the signature
        `f(params, x, **kwargs)` where `params` is a `PyTree`, `x` is a `PyTree`,
        and `f` should also return a `PyTree`.

      trace_axes:
        output axes to trace the output kernel over, i.e. compute only the trace
        of the covariance along the respective pair of axes (one pair for each
        axis in `trace_axes`). This allows to save space and compute if you are
        only interested in the respective trace, but also improve approximation
        accuracy if you know that covariance along these pairs of axes converges
        to a `constant * identity matrix` in the limit of interest (e.g.
        infinite width or infinite `n_samples`). A common use case is the channel
        / feature / logit axis, since activation slices along such axis are i.i.d.
        and the respective covariance along the respective pair of axes indeed
        converges to a constant-diagonal matrix in the infinite width or infinite
        `n_samples` limit.
        Also related to "contracting dimensions" in XLA terms.
        (https://www.tensorflow.org/xla/operation_semantics#dotgeneral)

      diagonal_axes:
        output axes to diagonalize the output kernel over, i.e. compute only the
        diagonal of the covariance along the respective pair of axes (one pair for
        each axis in `diagonal_axes`). This allows to save space and compute, if
        off-diagonal values along these axes are not needed, but also improve
        approximation accuracy if their limiting value is known theoretically,
        e.g. if they vanish in the limit of interest (e.g. infinite
        width or infinite `n_samples`). If you further know that on-diagonal
        values converge to the same constant in your limit of interest, you should
        specify these axes in `trace_axes` instead, to save even more compute and
        gain even more accuracy. A common use case is computing the variance
        (instead of covariance) along certain axes.
        Also related to "batch dimensions" in XLA terms.
        (https://www.tensorflow.org/xla/operation_semantics#dotgeneral)

      vmap_axes:
        A triple of `(in_axes, out_axes, kwargs_axes)`
        passed to `vmap` to evaluate the empirical NTK in parallel ove these axes.
        Precisely, providing this argument implies that `f(params, x, **kwargs)`
        equals to a concatenation along `out_axes` of `f` applied to slices of
        `x` and `**kwargs` along `in_axes` and `kwargs_axes`. In other words, it
        certifies that `f` can be evaluated as a `vmap` with `out_axes=out_axes`
        over `x` (along `in_axes`) and those arguments in `**kwargs` that are
        present in `kwargs_axes.keys()` (along `kwargs_axes.values()`).

        For example if `_, f, _ = nt.stax.Aggregate()`, `f` is called via
        `f(params, x, pattern=pattern)`. By default, inputs `x`, patterns
        `pattern`, and outputs of `f` are all batched along the leading `0`
        dimension, and each output `f(params, x, pattern=pattern)[i]` only
        depends on the inputs `x[i]` and `pattern[i]`. In this case, we can
        pass `vmap_axes=(0, 0, dict(pattern=0)` to specify along which dimensions
        inputs, outputs, and keyword arguments are batched respectively.

        This allows us to evaluate Jacobians much more
        efficiently. If `vmap_axes` is not a triple, it is interpreted as
        `in_axes = out_axes = vmap_axes, kwargs_axes = {}`. For example a very
        common use case is `vmap_axes=0` for a neural network with leading (`0`)
        batch dimension, both for inputs and outputs, and no interactions between
        different elements of the batch (e.g. no BatchNorm, and, in the case of
        `nt.stax`, also no Dropout). However, if there is interaction between
        batch elements or no concept of a batch axis at all, `vmap_axes` must be
        set to `None`, to avoid wrong (and potentially silent) results.

    Returns:
      A function `ntk_fn` that computes the empirical ntk.
    """

    def sum_and_contract(fx, j1, j2):
        ndim = fx.ndim
        size = utils.size_at(fx, trace_axes)

        _diagonal_axes = utils.canonicalize_axis(diagonal_axes, ndim)
        _trace_axes = utils.canonicalize_axis(trace_axes, ndim)

        def contract(x, y):
            param_axes = list(range(x.ndim))[ndim:]
            contract_axes = _trace_axes + param_axes
            return _dot_general(x, y, contract_axes, _diagonal_axes) / size

        return jax.tree.reduce(operator.add, jax.tree.map(contract, j1, j2))

    def ntk_fn(
        x1: PyTree, x2: PyTree | None, params: PyTree, **apply_fn_kwargs
    ) -> jnp.ndarray:
        """Computes a single sample of the empirical NTK (jacobian outer product).

        Args:
          x1:
            first batch of inputs.

          x2:
            second batch of inputs. `x2=None` means `x2=x1`. `f(x2)` must have a
            matching shape with `f(x1)` on `trace_axes` and `diagonal_axes`.

          params:
            A `PyTree` of parameters about which we would like to compute the
            neural tangent kernel.

          **apply_fn_kwargs:
            keyword arguments passed to `apply_fn`. `apply_fn_kwargs` will be split
            into `apply_fn_kwargs1` and `apply_fn_kwargs2` by the `split_kwargs`
            function which will be passed to `apply_fn`. In particular, the rng key
            in `apply_fn_kwargs`, will be split into two different (if `x1!=x2`) or
            same (if `x1==x2`) rng keys. See the `_read_key` function for more
            details.

        Returns:
          A single sample of the empirical NTK. The shape of the kernel is "almost"
          `zip(f(x1).shape, f(x2).shape)` except for:
          1) `trace_axes` are absent as they are contracted over.
          2) `diagonal_axes` are present only once.
          All other axes are present twice.
        """
        args1, args2, fx1, fx2, fx_axis, keys, kw_axes, x_axis = _get_args(
            f, apply_fn_kwargs, params, vmap_axes, x1, x2
        )

        def j_fn(x, *args):
            _kwargs = {k: v for k, v in zip(keys, args)}
            fx = _get_f_params(f, x, x_axis, fx_axis, kw_axes, **_kwargs)
            jx = jax.jacobian(fx)(params)
            return jx

        if not utils.all_none(x_axis) or not utils.all_none(kw_axes):
            in_axes = [x_axis] + [kw_axes[k] if k in kw_axes else None for k in keys]
            j_fn = jax.vmap(j_fn, in_axes=in_axes, out_axes=fx_axis)

        j1 = j_fn(x1, *args1)
        j2 = j_fn(x2, *args2) if not utils.all_none(x2) else j1
        ntk = jax.tree.map(sum_and_contract, fx1, j1, j2)
        return ntk

    return ntk_fn


## utils


def _get_f_params(
    f: Callable,
    x: PyTree,
    x_axis: PyTree,
    fx_axis: PyTree,
    kw_axes: dict[str, PyTree],
    **apply_fn_kwargs,
) -> Callable[[PyTree], PyTree]:
    x = _expand_dims(x, x_axis)

    apply_fn_kwargs = {
        k: _expand_dims(v, kw_axes[k]) if k in kw_axes else v
        for k, v in apply_fn_kwargs.items()
    }

    def _f(p: PyTree) -> PyTree:
        fx = f(p, x, **apply_fn_kwargs)
        return _squeeze(fx, fx_axis)

    return _f


def _ndim(x: PyTree) -> PyTree:
    return jax.tree.map(lambda x: x.ndim, x)


def _mod(x: PyTree | None, y: PyTree) -> PyTree:
    if x is None:
        return None
    return jax.tree.map(operator.mod, x, y)


def _squeeze(x: PyTree, axis: PyTree | None) -> PyTree:
    if axis is None:
        return x

    def squeeze(x: jnp.ndarray, axis: None | int | tuple[int, ...]) -> jnp.ndarray:
        """`np.squeeze` analog working with 0-sized axes."""
        if isinstance(axis, int):
            axis = (axis,)

        non_zero_axes = tuple()
        shift = 0

        for a in sorted(axis):
            if x.shape[a - shift] == 0:
                new_shape = x.shape[:a] + x.shape[a + 1 :]
                if utils.size_at(new_shape) == 0:
                    x = x.reshape(new_shape)
                else:
                    x = jnp.zeros(new_shape, x.dtype)

                shift += 1
            else:
                non_zero_axes += (a - shift,)

        return jnp.squeeze(x, non_zero_axes)

    return jax.tree.map(squeeze, x, axis)


def _expand_dims_array(x: _ArrayOrShape, axis: int) -> _ArrayOrShape:
    def expand(x: jnp.ndarray) -> jnp.ndarray:
        return jnp.expand_dims(x, axis)

    if isinstance(x, ShapedArray):
        return jax.eval_shape(expand, x)

    if isinstance(x, jnp.ndarray):
        return expand(x)

    raise TypeError(type(x), x)


def _expand_dims(
    x: None | PyTree | UndefinedPrimal, axis: PyTree | None
) -> PyTree | None:
    if axis is None or x is None or isinstance(x, UndefinedPrimal):
        return x
    return jax.tree.map(_expand_dims_array, x, axis)


def _get_args(
    f: Callable,
    apply_fn_kwargs: dict[str, PyTree],
    params: PyTree,
    vmap_axes: VMapAxes,
    x1: PyTree,
    x2: PyTree,
):
    kwargs1, kwargs2 = utils.split_kwargs(apply_fn_kwargs, x1, x2)

    fx1 = jax.eval_shape(f, params, x1, **kwargs1)
    fx2 = fx1 if utils.all_none(x2) else jax.eval_shape(f, params, x2, **kwargs2)

    x_axis, fx_axis, kw_axes = _canonicalize_axes(vmap_axes, x1, fx1, **kwargs1)

    keys = apply_fn_kwargs.keys()
    args1 = tuple(kwargs1[k] for k in keys)
    args2 = tuple(kwargs2[k] for k in keys)
    return args1, args2, fx1, fx2, fx_axis, keys, kw_axes, x_axis


def _canonicalize_axes(
    vmap_axes: VMapAxes | None, x: PyTree, fx: PyTree, **kwargs
) -> VMapAxisTriple:
    if isinstance(vmap_axes, tuple) and len(vmap_axes) == 3:
        x_axis, fx_axis, kw_axes = vmap_axes
    else:
        x_axis, fx_axis, kw_axes = vmap_axes, vmap_axes, {}

    if isinstance(x_axis, int):
        x_axis = jax.tree.map(lambda _: x_axis, x)

    if isinstance(fx_axis, int):
        fx_axis = jax.tree.map(lambda _: fx_axis, fx)

    if isinstance(kw_axes, int):
        kw_axes = jax.tree.map(lambda _: kw_axes, kwargs)

    x_axis = _mod(x_axis, _ndim(x))
    fx_axis = _mod(fx_axis, _ndim(fx))
    kw_axes = _mod(kw_axes, {k: _ndim(kwargs[k]) for k in kw_axes})
    return x_axis, fx_axis, kw_axes


def _get_res_batch_dims(
    contracting_dims: Iterable[int], batch_dims: Iterable[int]
) -> list[int]:
    res_batch_dims = [2 * b - i for i, b in enumerate(batch_dims)]
    for i, b in enumerate(batch_dims):
        for c in contracting_dims:
            if b > c:
                res_batch_dims[i] -= 2
    return res_batch_dims


def _dot_general(
    lhs: jnp.ndarray,
    rhs: jnp.ndarray,
    contracting_dims: Axes,
    batch_dims: Axes,
    precision=None,
) -> jnp.ndarray:
    """`jax.lax.dot_general` with preserved dims order and shared lhs / rhs dims.

    Precisely, returns `jax.lax.dot_general(lhs, rhs, dimension_numbers)` where
    `dimension_numbers == ((contracting_dims, contracting_dims),
                           (batch_dims, batch_dims))`,
    but preserves the dimension order in the output. See XLA's
     `DotGeneral<https://www.tensorflow.org/xla/operation_semantics#dotgeneral>`.

    Args:
      lhs: array.
      rhs: array, must have the same dimensionality as `lhs`.
      contracting_dims: contracting dimensions.
      batch_dims: batch dimensions.
      precision: Optional. Either `None`, which means the default precision for
        the backend, or a `Precision` enum value.

    Returns:
      Dot product result with preserved dimension order.
    """
    if lhs.ndim != rhs.ndim:
        raise ValueError(
            f"`lhs` and `rhs` must have the same dimensionality, got"
            f"`lhs.ndim == {lhs.ndim}` and `rhs.ndim == {rhs.ndim}`."
        )

    contracting_dims = utils.canonicalize_axis(contracting_dims, lhs)
    batch_dims = utils.canonicalize_axis(batch_dims, lhs)

    n_batch_dims = len(batch_dims)
    leading_batch_dims = range(n_batch_dims)

    dimension_numbers = ((contracting_dims, contracting_dims), (batch_dims, batch_dims))

    prod = jax.lax.dot_general(lhs, rhs, dimension_numbers, precision)
    prod = utils.zip_axes(prod, n_batch_dims)

    res_batch_dims = _get_res_batch_dims(contracting_dims, batch_dims)
    prod = jnp.moveaxis(prod, leading_batch_dims, res_batch_dims)
    return prod



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from functools import partial

import jax
import jax.numpy as jnp
from jax.sharding import PartitionSpec as P, NamedSharding

from netket.utils import config
from netket.utils.deprecation import warn_deprecation


safe_zip = partial(zip, strict=True)
_identity = lambda x: x


@partial(jax.jit, static_argnums=(0, 1))
def _prepare_mask(n, n_pad):
    return jnp.ones(n + n_pad, dtype=bool).at[-n_pad:].set(0)


def distribute_to_devices_along_axis(
    inp_data, axis=0, pad: bool = False, pad_value=None, devices=None
):
    """
    Distribute a local array equally along an axis to multiple jax devices devices

     .. note:
        Does nothing if netket.config.netket_experimental_sharding=False.

     .. note:
        Each jax process needs to have the whole array (parts not belonging to it can be filled with garbage).

    Args:
        inp_data: the full array (on every process)
        axis: (optional) axis along which to distribute
        pad: If True: pad the input data along axis to the next multiple of the number of devices
              If False (default): no padding; the size of the axis in inp_data needs to be divisible by the number of devices.
        pad_value: value to pad with (optional, only used if pad=True)
        devices: (optional) list of jax devices. Defaults to all available devices

    Returns:
        out_data: a distributed jax.Array
        mask: a mask indicating wether a given element is part of the original data (True) of of the padding (False)
              only returned if pad=True
    """
    if devices is None:
        devices = jax.devices()

    if config.netket_experimental_sharding:
        if pad:
            n = inp_data.shape[0]
            # pad to the next multiple of device_count
            device_count = jax.device_count()
            n_pad = math.ceil(inp_data.shape[0] / device_count) * device_count - n
            inp_data = jnp.pad(inp_data, ((0, n_pad), (0, 0)))
            if pad_value is not None and n_pad > 0:
                inp_data = inp_data.at[-n_pad:].set(pad_value)

        shape = [
            None,
        ] * inp_data.ndim
        shape[axis] = "S"
        mesh = jax.sharding.get_abstract_mesh()
        sharding = NamedSharding(mesh, P(*shape))
        out_data = jax.jit(_identity, out_shardings=sharding)(inp_data)

        if pad:
            if n_pad > 0:  # type: ignore
                mask = _prepare_mask(n, n_pad)  # type: ignore
            else:
                mask = None
            return out_data, mask
        else:
            return out_data
    else:
        return inp_data


# TODO consider merging this with distribute_to_devices_along_axis
@partial(jax.jit, static_argnames="axis")
def shard_along_axis(x, axis: int):
    """
    When running with experimental sharding mode, calls
    :func:`jax.lax.with_sharding_constraint` with a
    :class:`jax.sharding.NamedSharding` sharded along 'S' the given axis.

    Args:
        x: An array
        axis: the axis to be sharded
    """
    if config.netket_experimental_sharding and jax.device_count() > 1:
        # Shard shape is (1, 1, 1, -1, 1, 1) where -1 is the axis
        shard_shape = [None for _ in range(x.ndim)]
        shard_shape[axis] = "S"

        mesh = jax.sharding.get_abstract_mesh()
        x = jax.lax.with_sharding_constraint(
            x,
            NamedSharding(mesh, P(*shard_shape)),
        )
    return x


@jax.jit
def with_samples_sharding_constraint(x, shape=None):
    """
    ensure the input x is sharded along axis 0 on all devices
    works both outside and inside of jit
    """
    warn_deprecation(
        "with_samples_sharding_constraint is deprecated in favour of nk.jax.sharding.shard_along_axis(x, axis=0)"
    )

    return shard_along_axis(x, 0)


def extract_replicated(t):
    """
    Extract the value of a fully replicated global device array.

    Args:
        t: a jax Array (or a pytree of jax Arrays)

    Returns:
        A locally adressable representation of t
    """

    def _extract_replicated(x):
        if isinstance(x, jax.Array) and not x.is_fully_addressable:
            if not x.is_fully_replicated:
                raise RuntimeError(
                    "Expected a fully replicated array, but found one that is not. You should gather the array first. If you are not developing custom logic, please open a bug report with a reproducer."
                )
            return x.addressable_data(0)
        else:
            return x

    return jax.tree_util.tree_map(_extract_replicated, t)


def gather(x):
    """
    Make a sharded array fully replicated by gathering all parts on every device.

    Args:
        x: potentially unreplicated jax.Array with NamedSharding(...'S')

    Returns:
        fully replicated array
    """

    if not isinstance(x, jax.Array):
        # TODO in the future we could chagne it to just return x unchanged
        # but for now we error if x is not a jax array to ensure gather is used correctly
        raise RuntimeError("gather can only be applied to a jax.Array")
    elif x.is_fully_replicated:  # includes SingleDeviceSharding
        return x
    elif isinstance(x.sharding, NamedSharding):
        # x.sharding.device_set has arbitrary order
        # Hardcode all devices until I figure out a way to deduce the order from x
        out_shardings = NamedSharding(jax.sharding.get_abstract_mesh(), P())
    else:
        raise NotImplementedError(
            "Gather is not compatible with {x.sharding}. Please open a feature request."
        )
    return jax.jit(_identity, out_shardings=out_shardings)(x)



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from netket.utils.group import PermutationGroup, Identity, Permutation, Element
from netket.hilbert import DiscreteHilbert, SpinOrbitalFermions


def _physical_to_fermionic_permutation(
    perm: Element, hilbert: SpinOrbitalFermions
) -> Element:
    """Converts a permutation of the lattice sites to a permutation of single-particles states
    in a fermionic Hilbert space."""

    assert isinstance(perm, Permutation)

    perm_array = perm.permutation_array
    fermionic_perm_array = np.copy(perm_array)

    offset = 0
    while offset < hilbert.size - hilbert.n_orbitals:
        offset += hilbert.n_orbitals
        fermionic_perm_array = np.concatenate(
            (fermionic_perm_array, perm_array + offset)
        )

    return Permutation(permutation_array=fermionic_perm_array)


def physical_to_logical_permutation_group(
    perm_group: PermutationGroup, hilbert: DiscreteHilbert
) -> PermutationGroup:
    """Converts a permutation group of the lattice sites to a permutation group
    of the local degrees of freedom on the many-body Hilbert space."""

    if isinstance(hilbert, SpinOrbitalFermions):
        id_perm = np.arange(hilbert.size)
        fermionic_perms = []

        for perm in perm_group:
            if isinstance(perm, Identity):
                fermionic_perms.append(Permutation(permutation_array=id_perm))

            else:
                fermionic_perm = _physical_to_fermionic_permutation(perm, hilbert)
                fermionic_perms.append(fermionic_perm)

        perm_group = PermutationGroup(fermionic_perms, degree=hilbert.size)

    return perm_group



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netket.hilbert import AbstractHilbert, Qubit, Spin, Fock, SpinOrbitalFermions
from netket.utils.group import Permutation

from netket._src.operator.permutation.spin import (
    PermutationOperator,
)
from netket._src.operator.permutation.fermion import (
    PermutationOperatorFermion,
)


def construct_permutation_operator(
    hilbert_space: AbstractHilbert, permutation: Permutation
):
    """
    Return the appropriate permutation operator depending on the type of Hilbert space.

    If the Hilbert space is a spin or boson Hilbert space, a PermutationOperator will be returned,
    if it is a fermion Hilbert space, a PermutationOperatorFermion will be returned.

    For mathematical details on permutation operators, see :doc:`/advanced/symmetry`.

    Args:
        hilbert_space: The Hilbert space on which the permutation acts.
        permutation: The permutation to be represented as an operator.

    Returns:
        Either a PermutationOperator or PermutationOperatorFermion depending on the Hilbert space type.

    Raises:
        TypeError: If the Hilbert space type is not supported.
    """

    if isinstance(hilbert_space, (Qubit, Spin, Fock)):
        return PermutationOperator(hilbert_space, permutation)

    elif isinstance(hilbert_space, SpinOrbitalFermions):
        return PermutationOperatorFermion(hilbert_space, permutation)

    else:
        raise TypeError(
            "hilbert_space should be a Hilbert space of one of the following types:"
            "Qubit, Spin, Fock, SpinOrbitalFermions"
        )



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
from flax import nnx


class ApplyOperatorModuleNNX(nnx.Module):
    """
    A Flax NNX module that wraps another NNX module and applies an operator transformation.

    This module wraps a base neural network module and applies an operator in front of it,
    computing log(O|ψ⟩) where O is the operator and |ψ⟩ is represented by the base module.

    The operator is stored as a regular NNX Variable with collection='operator', which makes
    it non-trainable by default (since optimizers typically only update 'params' collection).

    Unlike the Linen version, NNX modules are stateful and contain their parameters directly.
    This makes the implementation more straightforward - we just store the base module and
    operator as attributes.

    Args:
        base_module: The NNX module to wrap
        operator: The operator to apply

    Example::

        import netket as nk
        from netket.models import RBM
        from flax import nnx

        # Create base NNX module (already initialized with parameters)
        base_module = RBM(N=10, alpha=2, rngs=nnx.Rngs(0))
        operator = nk.operator.spin.sigmax(hilbert, 0)

        # Create transformed module
        transformed = ApplyOperatorModuleNNX(base_module, operator)

        # Use it directly (NNX style)
        logpsi = transformed(x)

        # Or use with MCState
        vstate = nk.vqs.MCState(sampler, transformed, n_samples=1000)

        # The operator can be updated
        transformed.operator = new_operator
        logpsi = transformed(x)
    """

    def __init__(self, base_module: nnx.Module, operator):
        """
        Initialize the transformed module.

        Args:
            base_module: The base NNX module to wrap
            operator: The operator to apply in front of the base module
        """
        self.base_module = base_module

        # Cannot store directly the operator, as it would break the
        # assumption that variables are only pure dictionaries.
        leaves, treedef = jax.tree.flatten(operator)
        self._operator_treedef = treedef
        self._operator_leaves = nnx.Variable(leaves, collection="operator")

    @property
    def operator(self):
        """The operator applied to the base module."""
        return jax.tree.unflatten(
            self._operator_treedef, self._operator_leaves.get_value()
        )

    @operator.setter
    def operator(self, new_operator):
        leaves, treedef = jax.tree.flatten(new_operator)
        self._operator_treedef = treedef
        self._operator_leaves.set_value(leaves)

    def __nnx_repr__(self):
        """Custom repr to avoid issues with operator repr in NNX."""
        from flax.nnx import reprlib

        yield reprlib.Object(type(self))
        yield reprlib.Attr("base_module", self.base_module)
        yield reprlib.Attr("operator", self.operator)

    def __call__(self, x, *args, **kwargs):
        from netket.operator import ContinuousOperator

        # Get the operator (reconstructed from flattened representation)
        stored_operator = self.operator

        # Apply the operator transformation
        if isinstance(stored_operator, ContinuousOperator):
            # For continuous operators, we need to create a wrapper function
            # that extracts and recomposes the base module for the kernel
            def apply_base(variables, x_in, *args_in, **kwargs_in):
                # In NNX, we need to split and merge the base module
                graphdef, state = nnx.split(self.base_module)
                base_module_copy = nnx.merge(graphdef, variables)
                return base_module_copy(x_in, *args_in, **kwargs_in)

            # Split the base module to get its state as variables
            _, base_variables = nnx.split(self.base_module)

            res = stored_operator._expect_kernel(
                apply_base, base_variables, x, *args, **kwargs
            )
        else:
            # For discrete operators, use the standard approach
            xp, mels = stored_operator.get_conn_padded(x)
            xp = xp.reshape(-1, x.shape[-1])

            # Call the base module directly (NNX modules are stateful)
            logpsi_xp = self.base_module(xp, *args, **kwargs)
            logpsi_xp = logpsi_xp.reshape(mels.shape).astype(complex)

            # Compute log(sum(mels * exp(logpsi_xp)))
            res = jax.scipy.special.logsumexp(logpsi_xp, axis=-1, b=mels)

        return res



# Copyright 2021-2022 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, SupportsFloat, SupportsComplex, TYPE_CHECKING

import numbers
import warnings

import numpy as np
import jax.numpy as jnp
from scipy.sparse import issparse

from netket.hilbert import AbstractHilbert
from netket.utils.types import DType, Array
from netket.utils.numbers import dtype as _dtype, is_scalar
from netket.errors import OperatorMultiplicationDeprecationWarning

from .._discrete_operator import DiscreteOperator
from .._lazy import Transpose

from .helpers import (
    canonicalize_input,
    _multiply_operators,
    cast_operator_matrix_dtype,
)
from .convert import local_operators_to_pauli_strings

if TYPE_CHECKING:
    from .._pauli_strings import PauliStrings


def is_hermitian(a: np.ndarray, rtol=1e-05, atol=1e-08) -> bool:
    if issparse(a):
        return np.allclose(a.todense(), a.T.conj().todense(), rtol=rtol, atol=atol)
    else:
        return np.allclose(a, a.T.conj(), rtol=rtol, atol=atol)


def _is_sorted(a):
    for i in range(len(a) - 1):
        if a[i + 1] < a[i]:
            return False
    return True


class LocalOperatorBase(DiscreteOperator):
    """Base implementation of an operator composed of a sum of
    local terms, each of which acts on a small number of sites.

    .. warning::

        The complexity of the constructor of this operator scales
        exponentially with the number of sites on which each term
        acts, and linearly with the number of terms.

        The memory requirement scales the same, because this operator
        stores internally a lookup table as large as the local hilbert
        space size of each term.

        Refrain from using this for terms acting on more than
        6-sites. For Spin-1/2 systems, prefer instead
        :class:`netket.operator.PauliStrings`. For non spin-1/2
        there is nothing that will work efficiently out of the box,
        but you can easily roll your own.

    The runtime complexity is proportional to the number of
    connected entries
    :attr:`~netket.operator.DiscreteOperator.max_conn_size`
    """

    def __init__(
        self,
        hilbert: AbstractHilbert,
        operators: list[Array] | Array = [],
        acting_on: list[int] | list[list[int]] = [],
        constant: SupportsComplex | SupportsFloat = 0,
        dtype: DType | None = None,
        *,
        mel_cutoff: float = 1.0e-10,
    ):
        r"""
        Constructs a new ``LocalOperator`` given a hilbert space and (if
        specified) a constant level shift.

        Args:
           hilbert: Hilbert space the operator acts on.
           operators: A list of operators, in matrix form. Supports numpy dense or scipy
           sparse format
           acting_on: A list of list of sites, which the corresponding operators act on. This
                should be constructed such that :code:`operators[i]` acts on the sites :code:`acting_on[i]`.
                If operators is not a list of operators, acting_on should just be the list of
                corresponding sites.
           constant: Constant diagonal shift of the operator, equivalent to
                :math:`+\text{c}\hat{I}`. Default is 0.0.
           dtype: The datatype to use for the matrix elements. Defaults to double precision if
                available.
           mel_cutoff (float): a cutoff to remove small matrix elements (default = 1e-10)

        Examples:
           Constructs a ``LocalOperator`` without any operators.

           >>> from netket.hilbert import Spin
           >>> from netket.operator import LocalOperator
           >>> hi = Spin(0.5)**20
           >>> empty_hat = LocalOperator(hi)
           >>> print(len(empty_hat.acting_on))
           0
        """
        super().__init__(hilbert)
        self.mel_cutoff = mel_cutoff
        self._initialized = None
        self._is_hermitian = None

        # Canonicalize input. From now on input is guaranteed to be in canonical order
        operators, acting_on, dtype = canonicalize_input(
            self.hilbert, operators, acting_on, constant, dtype=dtype
        )
        self._dtype = dtype
        self._constant = np.array(constant, dtype=dtype)

        self._operators_dict = {}
        for op, aon in zip(operators, acting_on):
            self._add_operator(aon, op)

    def _add_operator(self, acting_on: tuple, operator: Array):
        """
        Adds an operator acting on a subset of sites.

        Does not modify in-place the operators themselves which are treated as
        immutables.
        """
        assert isinstance(acting_on, tuple)
        # acting_on_key = tuple(acting_on)
        if acting_on in self._operators_dict:
            operator = self._operators_dict[acting_on] + operator

        self._operators_dict[acting_on] = operator

    @property
    def operators(self) -> list[np.ndarray]:
        """List of the matrices of the operators encoded in this Local Operator.
        Returns a copy.
        """
        return list(self._operators_dict.values())

    @property
    def _operators(self) -> list[np.ndarray]:
        return self.operators

    @property
    def acting_on(self) -> list[list[int]]:
        """List containing the list of the sites on which every operator acts.

        Every operator `self.operators[i]` acts on the sites `self.acting_on[i]`
        """
        return list(self._operators_dict.keys())

    @property
    def n_operators(self) -> int:
        """The total number of operators that were summed upon to build this operator.

        This excludes an optional identity term that is tracked by
        :attr:`netket.operator.LocalOperator.constant`.
        """
        return len(self._operators_dict)

    @property
    def dtype(self) -> DType:
        """DType of the matrix elements of this operator."""
        return self._dtype

    @property
    def size(self) -> int:
        return self._size

    @property
    def is_hermitian(self) -> bool:
        """Returns true if this operator is hermitian."""
        # TODO: (VolodyaCO) I guess that if we have an operator with diagonal elements equal to 1j*C+Y, some complex constant, and
        # self._constant=-1j*C, then the actual diagonal would be Y. How do we check hermiticity taking into account the diagonal
        # elements as well as the self._constant? For the moment I just check hermiticity of the added constant, which must be real.
        if self._is_hermitian is None:
            self._is_hermitian = all(map(is_hermitian, self.operators)) and np.isreal(
                self._constant
            )

        return self._is_hermitian

    @property
    def mel_cutoff(self) -> float:
        r"""float: The cutoff for matrix elements.
        Only matrix elements such that abs(O(i,i))>mel_cutoff
        are considered"""
        return self._mel_cutoff

    @mel_cutoff.setter
    def mel_cutoff(self, mel_cutoff):
        self._mel_cutoff = mel_cutoff
        assert self.mel_cutoff >= 0

    @property
    def constant(self) -> numbers.Number:
        """
        A constant multiplying the identity added to the diagonal of the operator.
        """
        return self._constant

    def to_pauli_strings(self, **kwargs) -> "PauliStrings":  # noqa: F821
        """Convert to PauliStrings object"""
        return local_operators_to_pauli_strings(
            self.hilbert,
            self.operators,
            self.acting_on,
            self.constant,
            self.dtype,
            **kwargs,
        )

    def copy(self, *, dtype: DType | None = None, _cls=None):
        """Returns a copy of the operator, while optionally changing the dtype
        of the operator.

        Args:
            dtype: optional dtype

        Internal args:
            _cls: used to specify the target class
        """

        if dtype is None:
            dtype = self.dtype

        if not np.can_cast(self.dtype, dtype, casting="same_kind"):
            raise ValueError(f"Cannot cast {self.dtype} to {dtype}")

        if _cls is None:
            _cls = type(self)

        new = _cls(self.hilbert, constant=self.constant, dtype=dtype)
        new.mel_cutoff = self.mel_cutoff

        if dtype == self.dtype:
            new._operators_dict = self._operators_dict.copy()
        else:
            new._operators_dict = {
                aon: cast_operator_matrix_dtype(op, dtype)
                for aon, op in self._operators_dict.items()
            }

        return new

    def transpose(self, *, concrete=False):
        r"""LocalOperator: Returns the transpose of this operator."""
        if concrete:
            new = self.copy()
            for aon in new._operators_dict.keys():
                new._operators_dict[aon] = new._operators_dict[aon].transpose()
            return new
        else:
            return Transpose(self)

    def conjugate(self, *, concrete=False):
        r"""LocalOperator: Returns the complex conjugate of this operator."""
        new = self.copy()
        for aon in new._operators_dict.keys():
            new._operators_dict[aon] = new._operators_dict[aon].copy().conjugate()
        return new

    def __radd__(self, other):
        if is_scalar(other):
            return self.__add__(other)
        return super().__radd__(other)

    def __sub__(self, other):
        return self + (-other)

    def __rsub__(self, other):
        return other + (-self)

    def __isub__(self, other):
        return self.__iadd__(-other)

    def __neg__(self):
        return -1 * self

    def __add__(self, other: Union["LocalOperatorBase", numbers.Number]):
        if isinstance(other, LocalOperatorBase) or is_scalar(other):
            op = self.copy(dtype=jnp.promote_types(self.dtype, _dtype(other)))
            op = op.__iadd__(other)
            return op
        return super().__add__(other)

    def __iadd__(self, other):
        if isinstance(other, LocalOperatorBase):
            if self.hilbert != other.hilbert:
                return NotImplemented

            if not np.can_cast(other.dtype, self.dtype, casting="same_kind"):
                raise ValueError(
                    f"Cannot add inplace operator with dtype {other.dtype} "
                    f"to operator with dtype {self.dtype}"
                )

            self.mel_cutoff = min(other.mel_cutoff, self.mel_cutoff)
            self._constant += other.constant.item()
            for aon, op in other._operators_dict.items():
                self._add_operator(aon, op)

            self._reset_caches()
            return self
        if is_scalar(other):
            if not np.can_cast(type(other), self.dtype, casting="same_kind"):
                raise ValueError(
                    f"Cannot add inplace operator with dtype {type(other)} "
                    f"to operator with dtype {self.dtype}"
                )

            self._reset_caches()
            self._constant += other
            return self
        return NotImplemented

    def __truediv__(self, other):
        if not is_scalar(other):
            raise TypeError("Only division by a scalar number is supported.")

        if other == 0:
            raise ValueError("Dividing by 0")
        return self.__mul__(1.0 / other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __mul__(self, other):
        if isinstance(other, DiscreteOperator):
            # TODO: Deprecated in September 2025
            warnings.warn(OperatorMultiplicationDeprecationWarning())
            op = self.copy(dtype=jnp.promote_types(self.dtype, _dtype(other)))
            return op.__imatmul__(other)
        elif is_scalar(other):
            op = self.copy(dtype=jnp.promote_types(self.dtype, _dtype(other)))
            return op.__imul__(other)
        return NotImplemented

    def __imul__(self, other):
        if isinstance(other, DiscreteOperator):
            return self.__imatmul__(other)
        elif is_scalar(other):
            if not np.can_cast(_dtype(other), self.dtype, casting="same_kind"):
                raise ValueError(
                    f"Cannot add inplace operator of type {type(other)} and "
                    f"dtype {_dtype(other)} to operator with dtype {self.dtype}"
                )
            other = np.asarray(
                other, dtype=jnp.promote_types(self.dtype, _dtype(other))
            )

            self._constant *= other
            if np.abs(other) <= self.mel_cutoff:
                self._operators_dict = {}
            else:
                for key in self._operators_dict:
                    self._operators_dict[key] = other * self._operators_dict[key]

            self._reset_caches()
            return self

        return NotImplemented

    def __imatmul__(self, other):
        if not isinstance(other, LocalOperatorBase):
            return NotImplemented

        if not np.can_cast(other.dtype, self.dtype, casting="same_kind"):
            raise ValueError(
                f"Cannot add inplace operator with dtype {type(other)} to operator with dtype {self.dtype}"
            )

        return self._op_imatmul_(other)

    def _op__matmul__(self, other: "LocalOperatorBase") -> "LocalOperatorBase":
        if not isinstance(other, LocalOperatorBase):
            return NotImplemented
        op = self.copy(dtype=jnp.promote_types(self.dtype, _dtype(other)))
        return op._op_imatmul_(other)

    def _op_imatmul_(self, other: "LocalOperatorBase") -> "LocalOperatorBase":
        if not isinstance(other, LocalOperatorBase):
            return NotImplemented

        self.mel_cutoff = min(other.mel_cutoff, self.mel_cutoff)

        # (α + ∑ᵢAᵢ)(β + ∑ᵢBᵢ) =
        # = αβ + α ∑ᵢBᵢ + β ∑ᵢAᵢ + ∑ᵢⱼAᵢBⱼ
        # = β(α + ∑ᵢAᵢ) + α ∑ᵢBᵢ + ∑ᵢⱼAᵢBⱼ

        α = self.constant.item()
        β = other.constant.item()
        # copy A dict because it is modified inplace in __imul__(β) and add_operators
        A_op_dict = self._operators_dict.copy()
        B_op_dict = other._operators_dict

        # αβ + β ∑ᵢAᵢ
        self.__imul__(β)

        # α ∑ᵢBᵢ
        if np.abs(α) > self.mel_cutoff:
            for aon, op in B_op_dict.items():
                self._add_operator(aon, α * op)

        # ∑ᵢⱼAᵢBⱼ
        for supp_A_i, A_i in A_op_dict.items():
            for supp_B_j, B_j in B_op_dict.items():
                self._add_operator(
                    *_multiply_operators(
                        self.hilbert, supp_A_i, A_i, supp_B_j, B_j, dtype=self.dtype
                    )
                )

        is_hermitian_A = self._is_hermitian
        is_hermitian_B = other._is_hermitian

        self._reset_caches()

        if is_hermitian_A and is_hermitian_B:
            self._is_hermitian = is_hermitian_A

        return self

    def _reset_caches(self):
        """
        Cleans the internal caches built on the operator.
        """
        self._initialized = False
        self._is_hermitian = None

    @property
    def max_conn_size(self) -> int:
        """The maximum number of non zero ⟨x|O|x'⟩ for every x."""
        self._setup()
        return self._max_conn_size

    def __repr__(self):
        ao = self.acting_on

        acting_str = f"acting_on={ao}"
        if len(acting_str) > 55:
            acting_str = f"#acting_on={len(ao)} locations"
        return f"{type(self).__name__}(dim={self.hilbert.size}, {acting_str}, constant={self.constant}, dtype={self.dtype})"
